unit QTurtle;

interface

uses
  SysUtils, Classes, Types, Qt, QControls, QForms, QGraphics, QPrinters;

const MaxTurtles = $FF;

type
  TTurtles = array[0..MaxTurtles] of
               record                       { private properties of a Turtle }
                 xscreen, yscreen: real;    { screen coordinates of a Turtle (0/0) upper left}
                 alfa: real;                { angle in degrees; north has alfa = 0 }
                 PenColor: TColor;          { PenColor of a Turtle }
                 PenWidth: Integer;         { PenWidth of a Turtle }
               end;

  TTurtle = class(TScrollingWidget)
  private
    FTurtle: TTurtles;
    FWriteRel: boolean;     { if true then  PrTxt outputs right to Turtles direction }
    FNum: Integer;                         { number of a Turtle }
    FBeam: Boolean;                        { enable beaming in color finding process }
    FTH: Integer;                          { Turtleheight }
    FPen: TPen;                            { Pen property }
    FBrush: TBrush;                        { Brush property}
    FGroundColor: TColor;                  { color actually under the Turtle }
    FBGColor: TColor;                      { background color setting }
    FFont: TFont;                          { Font property }
    FVisible: Boolean;                     { Turtle state affected by HT/ST }
    FPenDown: Boolean;                     { Turtle state affected by PU/PD }
    FBufferBitMap: TBitMap;                { buffers canvas for a restore on paint }
    FCanvasWidth: integer;
    FCanvasHeight: integer;
    FCanvas: TControlCanvas;
    function GetCanvas: TCanvas;           { gets Turtle canvas }
    function GetPosX: Integer;             { gets X-Position of the actual Turtle }
    function GetPosY: Integer;             { gets Y-Position of the actual Turtle }
    procedure SetTurtleNum(Value: Integer);{ initializes a new Turtle and makes it actual}
    procedure SetBG( Value: TColor);       { sets backgrond color for the actual Turtle}
    procedure SetFont( Value: TFont);      { sets Font }
    procedure DrawTurtle;                  { draws a Turtle}
    procedure DrawTurtleWin;               { draws the Turtle client rect } 
    procedure RealizeNewPos;
    procedure CenterScrollPos;
  protected
    //  Botschafts-Verarbeitung unter Linux:
    function  EventFilter(Sender: QObjectH; Event: QEventH): Boolean; override;
    //  procedure WMPaint(var message: TWMPaint); message WM_Paint;
    //    wird ersetzt durch die private Methode DrawTurtleWin (s.o.), die
    //    von der berschriebenen Version von EventFilter aufgerufen wird;
    //  procedure WMSize(var message: TWMSize); message WM_Size;
    //    wird ersetzt durch die folgende ChangeBounds - Prozedur:
    procedure ChangeBounds(ALeft, ATop, AWidth, AHeight: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer);override;
    procedure ChangeBrush(Sender: TObject);
    procedure ChangePen(Sender: TObject);
    procedure ChangeFont(Sender: TObject);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure HOME;  { = moveTo(0,0) & Falfa=0. Turtlecoordinates are relative
                     { to the startposition (0/0) = Homeposition of the Turtle}
    procedure CS;    { clear screen }
    procedure ST;    { show Turtle }
    procedure HT;    { hide Turtle }
    procedure PD;    { pen down }
    procedure PU;    { pen up }
    procedure FD( Value: real);         { step forward }
    procedure BK( Value: real);         { step backward }
    procedure LT( Value: real);         { left turn }
    procedure RT( Value: real);         { right turn }
    procedure SetHXY(u, v: real);       { set heading to location (u,v)}
    procedure MoveXY(u, v: real);       { move Turtle to location (u,v)}
    function  FoundColor(Var dist: Integer; alfa: real; Col: TColor): Boolean;
   { searches within the distance <dist> and the Turtle-Angle <alfa> the nearest
     <Col>-colored point; if found, then <dist> = distance of such a point }
    procedure Fill;                      { fill with brush.color}
    procedure PrTxt(Text: string);       { print text right to Turtle }
    procedure SavePic(FileName: string); { save picture (BitMap) }
    procedure LoadPic(FileName: string); { load picture (BitMap) }
    procedure PrintPic(Wmm: Integer);    { print picture; Wmm = width in mm on paper }
  published
    property align;
    property BeamEnabled: boolean read FBeam write FBeam default FALSE;
    property BGColor: TColor read FBGColor write SetBG;            { backgroundcolor }
    property Pen: TPen read FPen write FPen;
    property Brush: TBrush read FBrush write FBrush;
    property Canvas: TCanvas read GetCanvas;    { gets Turtles Canvas }
    property Color: TColor read FGroundColor;   { gets color under the Turtle }
    property Font: TFont read FFont write SetFont;
    property TurtleHeight: Integer read FTH write FTH default 12;
    property Num: integer read FNum write SetTurtleNum default 0;
     { Multi-Turtle property; up to 255 Turtles can be accessed by the number <Num>:
     assigning x>=0 to <Num> activates the Turtle number x and makes this Turtle actual
     assigning x<0 to <Num> deactivates the Turtle number x, default Turtle gets actual
     range of <Num>: -Maxturtles..MaxTurtles ( number of the default Turtle is 0) }
    property PosX: Integer read GetPosX;  { Turtle X- & Y-postiton; coordinates are ~}
    property PosY: Integer read GetPosY;  { ~relative to Homeposition(0/0);read only }
    property WriteRel: boolean read FWriteRel write FWriteRel default false;
    property CanvasWidth: integer read  FCanvasWidth;  { gets Canvas width }
    property CanvasHeight: integer read  FCanvasHeight;{ gets Canvas height }
    property OnMouseDown;         {adapted: X & Y made relative to Turtle-Origin(0/0)}
    property OnMouseUp;           {adapted: X & Y made relative to Turtle-Origin(0/0)}
    property OnMouseMove;         {adapted: X & Y made relative to Turtle-Origin(0/0)}
  end;


procedure Register;

implementation

Const pidiv180 = pi/180;
Var
    IsNew : Boolean;
    XHome, YHome,
    Rx, Ry: Integer;                             { rounded Fx, Fy }
    dx, dy: Integer;
    TurtleRect: TRect;
    IsActive: array[0..MaxTurtles] of boolean;{ stores Turtles activity states }
    SaveTurtle: TTurtles;                     { stores last Turtles settings }


constructor TTurtle.Create;
begin
  inherited Create(AOwner);
  FCanvasWidth:=Screen.Width;
  FCanvasHeight:=Screen.Height;
  TurtleRect:=Bounds(0,0, FCanvasWidth, FCanvasHeight);
  FTH:=12;
  width:=150;
  height:=150;
  FCanvas:=TControlCanvas.create;
  TControlCanvas(FCanvas).control:=self;
  FBufferBitMap:=TBitMap.create;
  FBufferBitMap.width:= FCanvasWidth;
  FBufferBitMap.height:= FCanvasheight;
  XHome:=FCanvasWidth div 2;
  YHome:=FCanvasHeight div 2;
  FFont:=FBufferBitMap.Canvas.Font;
  FBrush:=TBrush.Create;
  FPen:=TPen.Create;
  FBrush.OnChange:=ChangeBrush;
  FPen.OnChange:=ChangePen;
  FFont.OnChange:=ChangeFont;
  FBGColor:=clWhite;
  FGroundColor:=FBGColor;
  FPendown:=TRUE;
  FVisible:=TRUE;
  IsActive[0]:=TRUE;
end;

function TTurtle.EventFilter(Sender: QObjectH; Event: QEventH): Boolean;
  var
  EventType: QEventType;
  begin
  EventType := QEvent_type(Event);
  Result := inherited EventFilter(Sender, Event);
  if EventType = QEventType_Paint then
    DrawTurtleWin;
  end;

procedure TTurtle.DrawTurtleWin;
  Var i, actualNum: integer;
  begin
  dX:=XHome - clientwidth + HorzScrollBar.position;
  dY:=YHome - clientheight + VertScrollBar.position;
  if (ComponentState = [csDesigning]) or IsNew then begin CS; IsNew:=false; end;
  FCanvas.Draw(-dX,-dY, FBufferBitMap);
  actualNum:=FNum;
  for i:=0 to MaxTurtles do begin FNum:=i; DrawTurtle; end;
  FNum:=actualNum;
  end;

procedure TTurtle.CenterScrollPos;
begin
  HorzScrollbar.position:=clientwidth div 2;
  VertScrollbar.position:=clientheight div 2;
end;

procedure TTurtle.ChangeBounds(ALeft, ATop, AWidth, AHeight: Integer);
begin
  inherited ChangeBounds(ALeft, ATop, AWidth, AHeight);
  HorzScrollbar.range:=clientwidth + clientwidth;
  VertScrollbar.range:=clientheight + clientheight;
  CenterScrollPos;
end;


procedure TTurtle.DrawTurtle;
Var rsa,rca,a,u,v: real;
begin
  if FVisible and IsActive[FNum]
  then with FTurtle[FNum], FCanvas do
  begin
    a:=FTurtle[FNum].alfa*pidiv180;
    rsa:=FTH*sin(a);
    rca:=FTH*cos(a);
    u:=xscreen-dx;
    v:=yscreen-dy;
    Pen.Width:=PenWidth;
    Pen.Color:=FBGColor xor PenColor;
    Pen.mode:=pmXor;
    Polyline([Point(round(u+rsa),round(v-rca)), Point(round(u+rca),round(v+rsa)),
              Point(round(u-rca),round(v-rsa)), Point(round(u+rsa),round(v-rca))]);
    Pen.mode:=pmCopy;
  end; {with}
end;

procedure TTurtle.CS;                   {clear screen }
begin
  Home;
  with FBufferBitMap.Canvas do
  begin
    brush.color:=FBGColor;
    FillRect(TurtleRect);
    brush.assign(FBrush);
  end; {with}
  FCanvas.Draw(-dX,-dY, FBufferBitMap);
  Drawturtle;
end;

procedure TTurtle.home;
Var i, actualNum: integer;
    penstate: boolean;
begin
  actualNum:=FNum;
  penstate:=FPendown;
  if FPendown then pu;
  for i:=0 to MaxTurtles do
  if IsActive[FNum] then begin
     FNum:=i;
     MoveXY(0,0); LT(FTurtle[FNum].alfa);
  end;
  FNum:=actualNum;
  FPendown:=penstate;
  CenterScrollPos;
end;

procedure TTurtle.RealizeNewPos;
  var LastPenPos : TPoint;
  begin
  with FCanvas do begin
    pen.color:=FTurtle[FNum].pencolor;
    pen.width:=FTurtle[FNum].penwidth;
    LastPenPos := Point(Round(SaveTurtle[FNum].xscreen),
                        Round(SaveTurtle[FNum].yscreen));
    if FPenDown then begin
      MoveTo(LastPenPos.X - dx,
             LastPenPos.Y - dy);
      LineTo(Rx - dx, Ry - dy);
      Pixels[LastPenPos.X - dx,
             LastPenPos.Y - dy] := pen.Color;
      Pixels[Rx - dx,  Ry - dy] := pen.Color;
      end
    else MoveTo(Rx-dx, Ry-dy);
  end; {with}
  with FBufferBitMap.Canvas do
  begin
    pen.color:=FTurtle[FNum].pencolor;
    pen.width:=FTurtle[FNum].penwidth;
    if FPenDown then begin
      MoveTo(LastPenPos.X,
             LastPenPos.Y);
      LineTo(Rx, Ry);
      Pixels[LastPenPos.X,
             LastPenPos.Y] := pen.Color;
      Pixels[Rx, Ry]       := pen.Color;
      end
    else MoveTo(Rx, Ry);
    FGroundColor := pixels[Rx, Ry];
  end; {with}
end;

procedure TTurtle.MoveXY;
begin
  DrawTurtle;
  with FTurtle[FNum] do begin
    SaveTurtle[FNum].xscreen:=xscreen;
    SaveTurtle[FNum].yscreen:=yscreen;
    xscreen:=XHome + u;
    yscreen:=YHome - v;
    Rx:=round(xscreen);
    Ry:=round(yscreen);
  end;
  RealizeNewPos;
  DrawTurtle;
end;

procedure TTurtle.SetHXY;
Var   dX,dY: real;
begin
  DrawTurtle;
  with FTurtle[FNum] do begin
       SaveTurtle[FNum].alfa:=alfa;
       dX:=(XHome + u) - xscreen;
       dY:=yscreen - (YHome - v);
       if dY>0
       then alfa:=ArcTan(dX/dY)/pidiv180
       else if dY<0
            then alfa:=180 + ArcTan(dX/dY)/pidiv180
            else if dX>0
                 then alfa:=90
                 else if dX<0
                      then alfa:=-90;
  end; {with}
  DrawTurtle;
 end;

procedure TTurtle.Fill;
  // Es wird ein nachbarschaftsorientierter Fll-Algorithmus benutzt:
  // Zwei Pixel heien benachbart, wenn sie eine gemeinsame Kante
  // besitzen. Der Nachteil, dass solche Algorithmen einzelne "in Ecken
  // isolierte" Pixel ungefllt lassen knnen, wird in Kauf genommen.
  // Jedes umzufrbende Pixel lst die Untersuchung seiner 4 Nachbarn
  // aus, ob diese ebenfalls umgefrbt werden mssen.
  // Fr einen sicheren Betrieb muss die zu fllende Flche die
  // folgenden Voraussetzungen erfllen:
  //   1. Sie muss vollstndig innerhalb des Turtle-Fensters liegen.
  //   2. Ihr Inneres muss komplett mit der Farbe BGColor gefllt sein.
  //   3. Ihr Rand muss eine geschlossene Linie sein.
  //   4. Die Farbe des Randes muss von BGColor verschieden sein.
{
  // Erster Entwurf:
  // Stark rekursive Fllprozedur, einfach, aber nicht effizient;
  //    fhrt schon bei mittelgroen Flchen zum Stack-berlauf !
  procedure floodfill(sx, sy: Integer);
    begin
    With FBufferBitMap do
    If (sx >= 0) and (sx < Width) and
       (sy >= 0) and (sy < Height) then
      with FBufferBitMap.Canvas do
        If Pixels[sx, sy] = BGColor then begin
          Pixels[sx, sy] := FBrush.Color;
          floodfill(sx + 1, sy);
          floodfill(sx, sy + 1);
          floodfill(sx - 1, sy);
          floodfill(sx, sy - 1);
          end;
    end;
}
  // Zweite Variante:
  // Rekursiver zeilenorientierter Fll-Algorithmus;
  //     deutlich weniger Rekursionsaufrufe, aber
  //     immmer noch recht langsam.
  procedure floodfill(sx, sy: Integer);
    // Rekursiver Kern:
    procedure ff(xs, ys, fxmin, fxmax, fy: Integer);
      var mxmin, mxmax : Integer;
      begin
      FBufferBitMap.Canvas.Pixels[xs, ys] := FBrush.Color;  // Startpixel frben
      mxmin := xs;
      While FBufferBitMap.Canvas.Pixels[mxmin - 1, ys] = BGColor do begin
        FBufferBitMap.Canvas.Pixels[mxmin - 1, ys] := FBrush.Color;
        mxmin := mxmin - 1;                         // Nach links ausfllen
        end;
      mxmax := xs;
      While FBufferBitMap.Canvas.Pixels[mxmax + 1, ys] = BGColor do begin
        FBufferBitMap.Canvas.Pixels[mxmax + 1, ys] := FBrush.Color;
        mxmax := mxmax + 1;                         // Nach rechts ausfllen
        end;
      DrawTurtleWin;
      xs := mxmin;
      Repeat                   // Oben nach Hintergrund-Pixeln suchen
        If ys - 1 = fy then    // Betrachtete Zeile ist schon bearbeitet
          While ((xs >= fxmin) and (xs <= fxmax)) or
                (FBufferBitMap.Canvas.Pixels[xs, ys - 1] <> BGColor) and
                (xs <= mxmax) do
            xs := xs + 1
        else                   // Neue Zeile
          While (FBufferBitMap.Canvas.Pixels[xs, ys - 1] <> BGColor) and
                (xs <= mxmax) do
            xs := xs + 1;
        If xs <= mxmax then begin
          ff(xs, ys - 1, mxmin, mxmax, ys); // Aufruf: Zeile drber bearbeiten
          xs := xs + 1;
          end;
      until xs > mxmax;
      xs := mxmin;
      Repeat                   // Unten nach Hintergrund-Pixeln suchen
        If ys + 1 = fy then    // Betrachtete Zeile ist schon bearbeitet
          While ((xs >= fxmin) and (xs <= fxmax)) or
                (FBufferBitMap.Canvas.Pixels[xs, ys + 1] <> BGColor) and
                (xs <= mxmax) do
            xs := xs + 1
        else                   // Neue Zeile
          While (FBufferBitMap.Canvas.Pixels[xs, ys + 1] <> BGColor) and
                (xs <= mxmax) do
            xs := xs + 1;
        If xs <= mxmax then begin
          ff(xs, ys + 1, mxmin, mxmax, ys); // Aufruf: Zeile drunter bearbeiten
          xs := xs + 1;
          end;
      until xs > mxmax;
      end; { of ff }

    var fxmin, fxmax, fy : Integer;  { Koordinaten der frisch gefllten Zeile }

    begin { of floodfill }
    If FBufferBitMap.Canvas.Pixels[sx, sy] = BGColor then begin
      FBufferBitMap.Canvas.Pixels[sx, sy] := FBrush.Color;  // Startpixel frben
      fxmin := sx;
      While FBufferBitMap.Canvas.Pixels[fxmin - 1, sy] = BGColor do begin
        FBufferBitMap.Canvas.Pixels[fxmin - 1, sy] := FBrush.Color;
        fxmin := fxmin - 1;                         // Nach links ausfllen
        end;
      fxmax := sx;
      While FBufferBitMap.Canvas.Pixels[fxmax + 1, sy] = BGColor do begin
        FBufferBitMap.Canvas.Pixels[fxmax + 1, sy] := FBrush.Color;
        fxmax := fxmax + 1;                         // Nach rechts ausfllen
        end;
      fy := sy;
      sx := fxmin;
      sy := fy + 1;
      While (FBufferBitMap.Canvas.Pixels[sx, sy] <> BGColor) and
            (sx <= fxmax) do
        sx := sx + 1;                 // Oben nach Hintergrund-Pixeln suchen
      If sx <= fxmax then
        ff(sx, sy, fxmin, fxmax, fy);     // Aufruf: Zeile drber bearbeiten
      sx := fxmin;
      sy := fy - 1;
      While (FBufferBitMap.Canvas.Pixels[sx, sy] <> BGColor) and
            (sx <= fxmax) do
        sx := sx + 1;                 // Unten nach Hintergrund-Pixeln suchen
      If sx <= fxmax then
        ff(sx, sy, fxmin, fxmax, fy);     // Aufruf: Zeile drunter bearbeiten
      end;
    end; { of floodfill }

  begin
  DrawTurtle;
  floodfill(Rx, Ry);
  FCanvas.Draw(-dX,-dY, FBufferBitMap);
  DrawTurtle;
  end;

function TTurtle.FoundColor(Var dist: integer; alfa: real; Col: TColor): Boolean;
Var u, v, dx, dy, alfa_now, alfa_save: real;
    steps, D: Integer;

{sub}procedure beam;
     Var i, OldTH: integer;
         VState, PState: boolean;
      begin
       VState:=FVisible;
       PState:=FPendown;
       ST; PU;
       OldTH:=FTH;
       FTH:=FTH div 3;
       DrawTurtle;
       RT(alfa);
       for i:=1 to steps*8 do FD(1/8); BK(steps);
       LT(alfa);
       DrawTurtle;
       FTH:=OldTH;
       FVisible:=VState;
       FPendown:=PState;
     end; {sub}

begin
  D:=abs(dist);
  alfa_save:=FTurtle[FNum].alfa;
  alfa_now:=(alfa_save + alfa)*pidiv180;
  dx := sin(alfa_now);
  dy := cos(alfa_now);
  u := FTurtle[FNum].xscreen;
  v := FTurtle[FNum].yscreen;
  for steps:=0 to D do
  begin
    u := u + dx;  v := v - dy;
    if FBufferBitMap.Canvas.Pixels[round(u),round(v)]=Col then break
  end;
  dist:=steps;
  result:=steps < D;
  if FBeam then beam;
  FTurtle[FNum].alfa:=alfa_save;
end;

procedure TTurtle.FD;
Var a: real;
begin {FD}
  DrawTurtle;
  with FTurtle[FNum] do
  begin
    SaveTurtle[FNum].xscreen := xscreen;
    SaveTurtle[FNum].yscreen := yscreen;
    a:=alfa*pidiv180;
    xscreen:= xscreen + Value*sin(a);
    yscreen:= yscreen - Value*cos(a);
    Rx:=round(xscreen);  Ry:=round(yscreen);
  end; {with}
  RealizeNewPos;
  DrawTurtle;
end;

procedure TTurtle.LT;
Var alfa: real;
begin
   DrawTurtle;  {remove old Turtle}
   alfa:=FTurtle[FNum].alfa;
   SaveTurtle[FNum].alfa:=alfa;
   alfa:=alfa - Value;
   if alfa > +180 then alfa := alfa - 360 else
   if alfa < -180 then alfa := alfa + 360;
   FTurtle[FNum].alfa:=alfa;
   DrawTurtle;  {redraw new Turtle}
end;

procedure TTurtle.BK;
begin
  FD(-Value);
end;

procedure TTurtle.RT;
begin
  LT(-Value);
end;

function TTurtle.GetCanvas: TCanvas;
begin
  result:=FBufferBitMap.Canvas;
  invalidate;
end;

function TTurtle.GetPosX: Integer;
begin
  result:=round(FTurtle[Num].xscreen - XHome);
end;

function TTurtle.GetPosY: Integer;
begin
  result:=round(YHome - FTurtle[FNum].yscreen)
end;

procedure TTurtle.SetBG( Value: TColor);
begin
  FBGColor:=Value;
  CS;
end;

procedure TTurtle.SetFont;
begin
  FFont.assign(Value);
end;

procedure TTurtle.SetTurtleNum;
begin
  if abs(value)<=MaxTurtles
  then if Value >= 0
       then with FTurtle[Value] do begin
              FNum:=Value;
              if not IsActive[Value]
              then begin
                     IsActive[Value]:=True;
                     xscreen:=XHome;
                     yscreen:=YHome;
                   end;
            end
       else begin
              if IsActive[-value]
              then begin FNum:=-Value; HT end;
              IsActive[-value]:=FALSE;
              FNum:=0;
            end;
end;

procedure TTurtle.HT;    { hide turtle }
begin
  if FVisible then begin DrawTurtle; FVisible:=false; end;
end;

procedure TTurtle.ST;    { show turtle }
begin
  if not FVisible then begin FVisible:=true; DrawTurtle; end;
end;

procedure TTurtle.PD;  { pen down }
begin
  FPenDown:=true;
end;

procedure TTurtle.PU;  { pen up }
begin
  FPenDown:=false;
end;

procedure TTurtle.PrTxt(Text: string);
  Var a: integer;
      t: real;
//      N: string;

{sub} procedure Prt_Txt(C: TCanvas; x,y: integer);
      begin
        x:= round(x + 10*sin(t)); x:= Rx - x;
        y:= round(y - 10*cos(t)); y:= Ry - y;
        with C do begin
          Font.Color:=FFont.Color;
          Brush.Color:=BGColor;
          TextOut(x,y, Text);
        end;
{sub} end;

begin
  if NOT FWriteRel then a:=0
  else a:=round(FTurtle[Fnum].alfa);
  if FCanvas.Font.Name ='System' then t:=-0.8 else t:=(a-60)*pidiv180;
  Prt_Txt(FBufferBitMap.Canvas,0,0);
  Prt_Txt(FCanvas,dx,dy);
end;

procedure TTurtle.SavePic;
begin
  DrawTurtle;   {remove}
  FBufferBitMap.SaveToFile(FileName);
  DrawTurtle;   {redraw}
end;

procedure TTurtle.LoadPic(FileName: string);
Var BMP: TBitMap;
begin
  BMP:=TBitMap.Create;
  with BMP do
  begin
    LoadFromFile(FileName);
    FBufferBitMap.Canvas.StretchDraw(TurtleRect, BMP);
    free;
  end;
  FBGColor:=FBufferBitMap.Canvas.Pixels[2,2];
  invalidate;
end;


procedure TTurtle.PrintPic(Wmm: Integer);
Var w,h: integer;
    ReDest: TRect;
begin
with printer do begin
  BeginDoc;
//  SetMapMode(Handle,MM_TEXT);
  w:=Wmm;
  h:=w * height Div width;
  ReDest:=bounds((pagewidth-w) div 2,(pageheight-h) div 2,w,h);
  Canvas.CopyRect(ReDest,FCanvas,ClientRect);
  EndDoc;
 end;
end;


procedure TTurtle.ChangeBrush;
begin
  FCanvas.Brush.assign(FBrush);
  FBufferBitMap.Canvas.Brush.assign(FBrush);
end;

procedure TTurtle.ChangePen;
begin
  DrawTurtle;   { remove using old color }
  FTurtle[FNum].PenColor:=FPen.Color;
  FTurtle[FNum].PenWidth:=FPen.Width;
  DrawTurtle;   { redraw using new color }
end;

procedure TTurtle.ChangeFont;
begin
  FCanvas.Font.assign(FFont);
end;

procedure TTurtle.MouseMove;
begin
  inherited MouseMove(Shift, X - clientwidth + HorzScrollBar.position,
                             clientheight-VertScrollBar.position - Y );
end;

procedure TTurtle.MouseUp;
begin
  inherited MouseUp(Button, Shift, X - clientwidth + HorzScrollBar.position,
                                   clientheight-VertScrollBar.position - Y );
end;

procedure TTurtle.MouseDown;
begin
  inherited MouseDown(Button, Shift, X - clientwidth + HorzScrollBar.position,
                                     clientheight-VertScrollBar.position - Y );
end;

destructor TTurtle.destroy;
begin
  FPen.free;
  FBrush.free;
  FBufferBitMap.free;
  FCanvas.free;
  inherited destroy;
end;

{======== Registrierung =================}

procedure Register;
begin
  RegisterComponents('Beispiele', [TTurtle]);
end;

end.
